@@ -6,16 +6,18 @@ use std::path::PathBuf;
 
 use anyhow::Result;
 use futures::future::BoxFuture;
+use futures::Future;
 use futures::FutureExt;
 use tokio::io::AsyncWrite;
 use tokio::io::AsyncWriteExt;
 use tokio::sync::mpsc::UnboundedReceiver;
 use tokio::sync::mpsc::UnboundedSender;
+use tokio::task::JoinHandle;
 
 use crate::fs_util;
 
 #[derive(Clone)]
-pub struct EnvState {
+pub struct ShellState {
   /// Environment variables that should be passed down to sub commands
   /// and used when evaluating environment variables.
   env_vars: HashMap<String, String>,
@@ -23,20 +25,30 @@ pub struct EnvState {
   /// not passed down to any sub commands.
   shell_vars: HashMap<String, String>,
   cwd: PathBuf,
+  /// The current shell or subshell's stdout sender.
+  /// This is used by async commands in order to avoid
+  /// other steps waiting on the async commands to finish
+  /// their output.
+  shell_stdout_tx: ShellPipeSender,
 }
 
-impl EnvState {
-  pub fn new(env_vars: HashMap<String, String>, cwd: PathBuf) -> Self {
+impl ShellState {
+  pub fn new(
+    env_vars: HashMap<String, String>,
+    cwd: &Path,
+    shell_stdout_tx: ShellPipeSender,
+  ) -> Self {
     let mut result = Self {
       env_vars: Default::default(),
       shell_vars: Default::default(),
       cwd: PathBuf::new(),
+      shell_stdout_tx,
     };
     // ensure the data is normalized
     for (name, value) in env_vars {
       result.apply_env_var(&name, &value);
     }
-    result.set_cwd(&cwd);
+    result.set_cwd(cwd);
     result
   }
 
@@ -109,6 +121,14 @@ impl EnvState {
       }
     }
   }
+
+  pub fn shell_stdout_tx(&self) -> ShellPipeSender {
+    self.shell_stdout_tx.clone()
+  }
+
+  pub fn set_shell_stdout_tx(&mut self, tx: ShellPipeSender) {
+    self.shell_stdout_tx = tx;
+  }
 }
 
 #[derive(Debug, PartialEq)]
@@ -120,20 +140,33 @@ pub enum EnvChange {
   Cd(PathBuf),
 }
 
-#[derive(Debug, PartialEq)]
+#[derive(Debug)]
 pub enum ExecuteResult {
-  Exit,
-  Continue(i32, Vec<EnvChange>),
+  Exit(i32, Vec<JoinHandle<()>>),
+  Continue(i32, Vec<EnvChange>, Vec<JoinHandle<()>>),
 }
 
-pub struct ExecutedTask {
+impl ExecuteResult {
+  pub fn into_handles(self) -> Vec<JoinHandle<()>> {
+    match self {
+      ExecuteResult::Exit(_, handles) => handles,
+      ExecuteResult::Continue(_, _, handles) => handles,
+    }
+  }
+}
+
+pub struct SpawnedStep {
   pub stdout: ShellPipe,
   pub task: BoxFuture<'static, ExecuteResult>,
 }
 
-impl ExecutedTask {
+impl SpawnedStep {
   pub fn from_exit_code(exit_code: i32) -> Self {
-    Self::from_result(ExecuteResult::Continue(exit_code, Vec::new()))
+    Self::from_result(ExecuteResult::Continue(
+      exit_code,
+      Vec::new(),
+      Vec::new(),
+    ))
   }
 
   pub fn from_result(execute_result: ExecuteResult) -> Self {
@@ -155,11 +188,45 @@ impl ExecutedTask {
       task: async move {
         let _ = tx.send(text.into_bytes());
         drop(tx); // close stdout
-        ExecuteResult::Continue(0, Vec::new())
+        ExecuteResult::Continue(0, Vec::new(), Vec::new())
       }
       .boxed(),
     }
   }
+
+  pub fn from_fn(
+    func: impl FnOnce() -> ExecuteResult + Sync + Send + 'static,
+  ) -> Self {
+    Self::from_future(async move { func() })
+  }
+
+  pub fn from_future(
+    future: impl Future<Output = ExecuteResult> + Send + 'static,
+  ) -> Self {
+    let (tx, stdout) = ShellPipe::channel();
+    SpawnedStep {
+      stdout,
+      task: async move {
+        let result = future.await;
+        drop(tx); // close stdout
+        result
+      }
+      .boxed(),
+    }
+  }
+
+  pub async fn wait_with_stdout<T>(
+    self,
+    action: impl FnOnce(ShellPipe) -> T,
+  ) -> ExecuteResult
+  where
+    T: Future<Output = ()> + Send + 'static,
+  {
+    let output_task = tokio::task::spawn(action(self.stdout));
+    let result = self.task.await;
+    output_task.await.unwrap();
+    result
+  }
 }
 
 pub type ShellPipeReceiver = UnboundedReceiver<Vec<u8>>;
@@ -179,25 +246,32 @@ impl ShellPipe {
     (data_tx, ShellPipe::Channel(data_rx))
   }
 
+  /// Unwraps the pipe to a channel.
+  pub fn unwrap_channel(self) -> ShellPipeReceiver {
+    match self {
+      ShellPipe::InheritStdin => unreachable!(),
+      ShellPipe::Channel(rx) => rx,
+    }
+  }
+
   /// Write everything to the specified writer
   pub async fn write_all(
     self,
-    mut writer: impl AsyncWrite + std::marker::Unpin,
+    mut writer: Box<dyn AsyncWrite + std::marker::Unpin + Send + Sync>,
   ) -> Result<()> {
-    match self {
-      ShellPipe::InheritStdin => unreachable!(),
-      ShellPipe::Channel(mut rx) => {
-        while let Some(data) = rx.recv().await {
-          writer.write(&data).await?;
-        }
-      }
+    let mut rx = self.unwrap_channel();
+    while let Some(data) = rx.recv().await {
+      writer.write(&data).await?;
     }
     Ok(())
   }
 
-  /// Pipes this pipe to the current process' stdout.
-  pub async fn pipe_to_stdout(self) {
-    let _ = self.write_all(tokio::io::stdout()).await;
+  /// Pipes this pipe to the specified writer.
+  pub async fn pipe_to_writer(
+    self,
+    writer: Box<dyn AsyncWrite + Unpin + Send + Sync>,
+  ) {
+    let _ = self.write_all(writer).await;
   }
 
   /// Pipes this pipe to the specified sender.
@@ -214,40 +288,3 @@ impl ShellPipe {
     }
   }
 }
-
-// #[cfg(test)]
-// mod test {
-//   use crate::parse_string_parts;
-
-//   use super::*;
-
-//   #[test]
-//   fn evaluate_string_parts_test() {
-//     run_evaluate_sp_test("$Test", &[], &[], "");
-//     run_evaluate_sp_test("$Test", &[("Test", "value")], &[], "value");
-//     run_evaluate_sp_test("$Test", &[], &[("Test", "value")], "value");
-//     run_evaluate_sp_test("$A a", &[], &[("A", "value")], "value a");
-//     run_evaluate_sp_test("\\$A a", &[], &[("A", "value")], "$A a");
-//   }
-
-//   fn run_evaluate_sp_test(
-//     text: &str,
-//     shell_vars: &[(&str, &str)],
-//     env_vars: &[(&str, &str)],
-//     expected: &str,
-//   ) {
-//     let (_, parts) = parse_string_parts(|_| true)(text).unwrap();
-//     let state = EnvState {
-//       shell_vars: shell_vars
-//         .iter()
-//         .map(|(k, v)| (k.to_string(), v.to_string()))
-//         .collect(),
-//       env_vars: env_vars
-//         .iter()
-//         .map(|(k, v)| (k.to_string(), v.to_string()))
-//         .collect(),
-//       cwd: PathBuf::from("/current_dir"),
-//     };
-//     assert_eq!(state.evaluate_string_parts(&parts), expected)
-//   }
-// }
